# W41 - TP noté

Vous êtes en charge du développement d'une API HTTP qui permet de manipuler des données relatives à une application de gestion de projets.

Cette API doit être développée en __Node.JS__ avec __Express__, en lien avec une base de données au travers de l'ORM __Prisma__.

> Vous devez travailler dans le dossier `tp_note_2023_2024` fourni.
> 
> N'oubliez pas de commencer par installer les dépendances :
> ```sh
> cd tp_note_2023_2024
> npm install
> ```

## Modèle de données

Le modèle de données est le suivant :

- `User`: un utilisateur de l'application
  - `email`: adresse email (string)
  - `username`: nom d'utilisateur (string - optionnel)
  - `createdProjects`: projets créés par l'utilisateur (références vers des `Project`)
  - `assignedTasks`: tâches assignées à l'utilisateur (références vers des `Task`)

- `Project`: un projet
  - `name`: nom du projet (string)
  - `description`: description du projet (string - optionnel)
  - `creator`: créateur du projet (référence vers un `User`)
  - `tasks`: tâches du projet (références vers des `Task`)

- `Task`: une tâche
  - `title`: titre de la tâche
  - `cost`: coût de la tâche (entier)
  - `dueDate`: date de fin de la tâche (date)
  - `completed`: indique si la tâche est terminée ou non (booléen)
  - `project`: projet auquel appartient la tâche (référence vers un `Project`)
  - `assignees`: utilisateurs assignés à la tâche (références vers des `User`)

Il y a donc les relations suivantes entre ces entités :

- entre `User` et `Project` : un utilisateur peut créer plusieurs projets, un projet n'a qu'un seul créateur
- entre `Project` et `Task` : un projet peut contenir plusieurs tâches, une tâche appartient à un seul projet
- entre `User` et `Task` : un utilisateur peut être assigné à plusieurs tâches, une tâche peut avoir plusieurs assignés

La suppression d'un `User` entraîne la suppression de tous les `Project` qu'il a créés.

La suppression d'un `Project` entraîne la suppression de toutes les `Task` associées.

> Décrire le modèle de données dans un schéma Prisma.

## API

### Données et codes de retour

Les routes de l'API doivent respecter les conventions suivantes :

- pour toutes les requêtes `POST` et `PATCH`, les données sont attendues au format JSON dans le corps de la requête
- les routes de récupération de ressource (`GET`) retournent les données au format JSON avec un status code 200
- les routes de création de ressource (`POST`) retournent la nouvelle entité créée au format JSON avec un status code 201
- les routes de modification de ressource (`PATCH`) retournent la ressource modifiée au format JSON avec un status code 200
- les routes de suppression de ressource (`DELETE`) retournent uniquement un status code 204
- les routes qui ciblent une ressource particulière (avec son identifiant) retournent un status code 404 si la ressource correspondante n'existe pas

### Validation des données client

Utiliser la bibliothèque `superstruct` pour valider les données reçues de la part du client dans le corps des requêtes `POST` et `PATCH` ou dans la query string.

Si une donnée est invalide, retourner un status code 400 avec un message d'erreur indiquant la raison de l'échec de la validation.

### Routes

L'API HTTP doit exposer les routes suivantes, que vous devez déclarer et implémenter dans une application `Express` :

- `GET /users` : retourne la liste des utilisateurs
- `GET /users/:user_id` : retourne les informations de l'utilisateur `:user_id`
- `POST /users` : crée un nouvel utilisateur
- `PATCH /users/:user_id` : modifie les informations de l'utilisateur `:user_id`
- `DELETE /users/:user_id` : supprime l'utilisateur `:user_id`

---

- `POST /users/:user_id/projects` : crée un nouveau projet dont l'utilisateur `:user_id` est le créateur
- `GET /users/:user_id/projects` : retourne la liste des projets créés par l'utilisateur `:user_id`

---

- `PATCH /projects/:project_id` : modifie les informations du projet `:project_id`
- `DELETE /projects/:project_id` : supprime le projet `:project_id`

---

- `POST /projects/:project_id/tasks` : crée une nouvelle tâche dans le projet `:project_id`
- `GET /projects/:project_id/tasks` : retourne la liste des tâches du projet `:project_id`

---

- `PATCH /tasks/:task_id` : modifie les informations de la tâche `:task_id`
- `DELETE /tasks/:task_id` : supprime la tâche `:task_id`

---

- `GET /tasks/:task_id/assignees` : retourne la liste des utilisateurs assignés à la tâche `:task_id`
- `POST /tasks/:task_id/assignees/:user_id` : assigne l'utilisateur `:user_id` à la tâche `:task_id` (en cas de succès, retourne uniquement un status code 200 et pas de données)
- `DELETE /tasks/:task_id/assignees/:user_id` : retire l'utilisateur `:user_id` de la tâche `:task_id` (en cas de succès, retourne uniquement un status code 200 et pas de données)

---

- `GET /users/:user_id/tasks` : retourne la liste des tâches assignées à l'utilisateur `:user_id`

Cette dernière route accepte également un paramètre optionnel `completed` dans la query string, dont les valeurs possibles sont "true" ou "false", et qui permet de filtrer les tâches en fonction de leur état de complétion.

Pour cette route, il faut également inclure les informations (`id` et `name`) du `Project` auquel chaque `Task` retournée appartient.

---

Un peu plus difficile :

- `GET /projects/:project_id/users` : retourne la liste des utilisateurs assignés à des tâches du projet `:project_id`

_Indication_ : une approche possible est de faire un `findMany` centré `User` en filtrant pour ne garder que ceux dont l'ensemble des tâches assignées en contient au moins une associée au projet `:project_id`.

- `GET /projects/:project_id/cost` : retourne le coût total des tâches du projet `:project_id`

_Indication_ : une approche possible est d'utiliser un `groupBy` pour regrouper les tâches par projet, filtrer pour ne garder que le projet `:project_id`, puis de demander la somme des `cost` (https://www.prisma.io/docs/orm/prisma-client/queries/aggregation-grouping-summarizing#group-by).
