import { prisma } from '../db';
import type { Request, Response } from 'express';
import { HttpError } from '../error';
import { assert } from 'superstruct';

import { UserCreationData, UserUpdateData } from "../validation/user";
import { responde } from "../functions";
import { ProjectCreationData } from '../validation/project';

export async function get_all(req: Request, res: Response) {
  const users = await prisma.user.findMany({
    include: {
      createdProjects: {
        include: {
          tasks: true,
        },
      },
    },
  });
  res.status(200).json(users);
}

export async function get_one(req: Request, res: Response) {
  const user = await prisma.user.findUnique({
    where: { id: Number(req.params.user_id) },
    include: {
      createdProjects: {
        include: {
          tasks: true,
        },
      },
    },
  });
  if (user == null) {
    throw new HttpError("User doesn't exist", 404)
  }
  res.status(200).json(user);
}

export async function create_one(req: Request, res: Response) {
  assert(req.body, UserCreationData);
  const user = await prisma.user.create({ data: req.body });
  responde(res, 201, "User succesfully created.", user);
}

export async function update_one(req: Request, res: Response) {
  assert(req.body, UserUpdateData);

  const user = await prisma.user.update({
    where: { id: Number(req.params.user_id) },
    data: req.body,
    include: {
      createdProjects: true,
    },
  });
  responde(res, 200, "User succesfully updated.", user);
}

export async function delete_one(req: Request, res: Response) {
  const user = await prisma.user.delete({ where: { id: Number(req.params.user_id) }});
  if (user == null) {
    throw new HttpError("User doesn't exist", 404)
  }
  responde(res, 204, "User succesfully deleted.");
}

export async function create_user_project(req: Request, res: Response) {
  assert(req.body, ProjectCreationData);
  const user = await prisma.user.update({
    where: {
      id: Number(req.params.user_id)
    },
    data: {
      createdProjects: {
        create: req.body
      }
    },
    include: {
      createdProjects: true,
    }
  });
  if (user == null) {
    throw new HttpError("User doesn't exist", 404);
  }
  responde(res, 201, "Project succesfully created and added to user", user);
}

export async function get_user_project(req: Request, res: Response) {
  const user = await prisma.user.findUnique({
    where: { id: Number(req.params.user_id) },
    include: {
      createdProjects: {
        include: {
          tasks: true
        }
      },
    },
  });
  if (user == null) {
    throw new HttpError("User doesn't exist", 404)
  }
  res.status(200).json(user.createdProjects);
}